// KSeqDemo.cpp : Defines the entry point for the application.
//

#include "stdafx.h"
#include "IOBase.h"
#include "kseq.h"

#define VCOUNT	4096
#define CCOUNT	128

/*
** Working copies of external memory. Use these for internal logic to minimize overhead of
** accessing protocol copy.
*/
WORD VMemory[VCOUNT];
BYTE CMemory[CCOUNT];

/*
** Functions to simplify access of working copies of external memory
*/
inline WORD GetV(int Address){ return VMemory[Address]; }
inline void SetV(int Address, WORD Val){ VMemory[Address] = Val; }

BOOL GetC(int Address)
{ 
	return (CMemory[Address/8] >> (Address%8)) & 0x01; 
}

void SetC(int Address, BOOL Val)
{
	int ByteAddr = Address/8;
	int ByteOffset = Address%8;
	BYTE CurVal = CMemory[ByteAddr] & ~(0x01<<ByteOffset);
	CMemory[ByteAddr] = CurVal | (Val ? (0x01<<ByteOffset) : 0);
}
		
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, int nCmdShow)
{
	OutputDebugString(L"Starting KSeqDemo");

	/*
	** Create and initialize base access class
	*/
	CIOBase IOB;
	IOB.Init();
	IOB.ReadIO();

	// This project expects an 8 point input in slot 0, it exits if it doesn't find one
	if(!IOB.GetModule(0))
		{
		OutputDebugString(L"No sim! Aborting...");
		return -1;
		}

	/*
	** Create K Sequence object and initialize
	*/
	KSeqHandler KSQ(VCOUNT, CCOUNT);

	// This initializes the handler as Ethernet on the specified IP port.
//	KSQ.InitEthernetHandler(0x7676); // Port shouldn't be 0x7070 or 0x7272, but anything else above 5000 is fine

	// This initializes the handler as serial on the specified comm port.
	KSQ.InitSerialHandler(L"COM2:", CBR_9600, 8, NOPARITY, ONESTOPBIT);

	/*
	** Primary scan loop, turning off input 0/0 stops scan
	*/
	while(IOB.GetDI(0, 0))
		{
		/*
		** Lock access to master data registers at beginning of scan. While locked, the protocol
		** handler cannot write new values but can service reads from its internal cache.
		*/
		KSQ.BeginBitAccess();
		KSQ.BeginWordAccess();

		/*
		** Refresh local copy of data from master copy
		*/
		KSQ.ReadWordMemory(0, VCOUNT, VMemory);
		KSQ.ReadBitMemory(0, CCOUNT, CMemory);

		/*
		** Normal PLC scan
		*/
		// Read IO
		IOB.ReadIO();

		// Solve logic
		DWORD TickCount = GetTickCount();
		SetV(0, LOWORD(TickCount));
		SetV(1, HIWORD(TickCount));
		SetV(2, GetV(0)/2);
		SetV(3, GetV(1)/2);

		TickCount /= 1000;
		SetC(0, TickCount & 0x01);
		SetC(1, TickCount & 0x02);
		SetC(2, TickCount & 0x04);
		SetC(3, TickCount & 0x08);
		SetC(4, TickCount & 0x10);
		SetC(5, TickCount & 0x20);
		SetC(6, TickCount & 0x40);
		SetC(7, TickCount & 0x80);

		// Write IO
		IOB.WriteIO();

		/*
		** Update master copy with local copy
		*/
		KSQ.WriteWordMemory(0, VCOUNT, VMemory);
		KSQ.WriteBitMemory(0, CCOUNT, CMemory);

		/*
		** Update handler's read cache. All protocol reads get their data from this cache so it must
		** be updated every scan to be accurate.
		*/
		KSQ.UpdateMemoryCache();

		/*
		** Unlock access to master data registers. This allows the protocol handler to access
		** the master data to complete write operations
		*/
		KSQ.EndBitAccess();
		KSQ.EndWordAccess();

		/*
		** Do post scan idle stuff. Optionally allow more time for the protocol handler thread
		** by sleeping here. The sleep value is arbitrary.
		*/
		Sleep(5);
		IOB.OnIdle();
		}

	/*
	** Clean up outputs and shut down base interface
	*/
	IOB.ClearOutputs();
	IOB.WriteIO();
	IOB.Exit();

	/*
	** Shut down protocol handler
	*/
	KSQ.ExitHandler();

	OutputDebugString(L"Exiting KSeqDemo");

	return 0;
}

